import uuid
from concurrent.futures import Future, ThreadPoolExecutor

class Agent:
    """
    Generally speaking, agents running in the system are instantiated from my subclasses. I will interpret the message I received. If I understand it, I will execute the semantics of the message, possibly replying asynchronously with the result of the calculation as needed;  If I do not understand the message, I reply with "Message Not Understood." If there is an error in interpreting the message along the way, I reply with an error message.
    """

    def __init__(self, id=None, receive_own_broadcasts=False, debugging=False) -> None:
        self._RESPONSE_ACTION_NAME = "[response]"
        self._ERROR_ACTION_NAME = "[error]"
        self.broadcastFlag = "[broadcast]"
        self._futures = {}
        self.availableActions = {}

        self.supervisor = None
        if id:
            self.id = id  # agent id
        else:
            self.id = self.__class__.__name__
        self.current_message = None

        # for debugging
        self.debugging = debugging
        self._logs = []

    def setSupervisor(self, supervisor):
        self.supervisor = supervisor

    def debugLog(self, something):
        # for debugging
        if self.debugging:
            self._logs.append(something)

    def clearLog(self):
        # for debugging
        self._logs = []

    def _receive(self, message) -> None:
        self.current_message = message
        self.interpret(self.current_message)

    def interpret(self, message):
        # The object interprets the message it understands
        agentName = message["action"]["name"]
        agentID = message["meta"]["id"]
        parent_id = message["meta"].get("parent_id")

        if agentName == self._RESPONSE_ACTION_NAME:
            #Handle incoming responses. Only useful when agent is used as callee
            # console.debug("Handle incoming response", message);
            # process help response
            args_value = message["action"]["args"]["value"]
            if (type(args_value) == dict and args_value.get("help")):
                self.availableActions[message["from"]] = args_value.get("help")
                
            # process _futures
            future = self._futures.get(parent_id)
            if future:
                future.set_result(message["action"]["args"]["value"])
                # del self._futures[parent_id]
        elif agentName == self._ERROR_ACTION_NAME:
            # console.debug("Handle incoming error", message);
            future = self._futures.get(parent_id)
            future.set_exception(Exception(message["action"]["args"]["error"]))
        else:
            # the caller requests the agent to execute the command
            self._commit(message)

    def _commit(self, message):
        try:
            # todo policy
            action_method = getattr(self, message["action"]["name"], None)
            if action_method:
                action_method(*message["action"]["args"])
            else:
                error = f'Message Not Understood: {message["to"]}>>{message["action"]["name"]}'
                self.raiseWith(error)
                return
        except Exception as e:
            error = f'{self.id}>>{message["action"]["name"]} raised exception:' + str(e)
            # console.error(error);
            self.raiseWith(error)

    def generateMessage(self, parent_id, to, action, args):
        # Translated from dynatalk-js by chatgpt
        message = {
            "meta": {
                "id": "",
            },
            "from": "",
            "to": "",
            "action": {
                "name": "",
                "args": "",
            },
        }
        
        if parent_id:
            message["meta"]["parent_id"] = parent_id
        message["meta"]["id"] = str(uuid.uuid4())
        message["to"] = to
        message["from"] = self.id
        message["action"]["name"] = action
        message["action"]["args"] = args
        
        return message

    def send(self, message):
        """
        Sends (out) a message from self agent.

        Args:
              message: The message

        Returns:
              The meta.id of the sent message
        """

        if self.debugging:
            self.debugLog(f"{self.id}: sending: {message}")
        # self._outbound_queue.put(message) // todo loop, now is directly
        self.supervisor.send(message)
        return message["meta"]["id"]

    def _request(self, message, timeout=3):
        # send and wait the response
        # timeout 3000 ms
        # The sync mechanism is implemented by the agent itself.
        with ThreadPoolExecutor() as executor:
            future = Future()
            task = executor.submit(lambda f: f.result(timeout=timeout), future)
            msg_id = self.send(message)
            self._futures[msg_id] = future  # todo: threadsafe
            return task.result()  # block and wait the result or TimeoutError

    def request(self, agentName, actionName, args):
        parent_id = None
        message = self.generateMessage(parent_id, agentName, actionName, args)
        return self._request(message)

    def sendTo(self, agentName, actionName, args):
        parent_id = None
        message = self.generateMessage(parent_id, agentName, actionName, args)
        return self.send(message)

    def respondWith(self, value):
        # Sends a response with the given value.
        parent_id = self.current_message["meta"]["id"]
        to = self.current_message["from"]
        action = self._RESPONSE_ACTION_NAME
        args = {"value": value}
        message = self.generateMessage(parent_id, to, action, args)
    
        self.send(message)

    def raiseWith(self, error):
        """
        Sends an error response.

        Args:
            error: The error to send.
        """
        parent_id = self.current_message["meta"]["id"]
        to = self.current_message["from"]
        action = self._ERROR_ACTION_NAME
        args = {"error": str(error)}
        message = self.generateMessage(parent_id, to, action, args)
    
        self.send(message)

    def ping(self):
        self.respondWith("pong")

    def broadcast(self, actionName, args):
        self.sendTo(self.broadcastFlag, actionName, args)

    def broadcastHelp(self):
        self.broadcast("help", [])

        

class PyDemoAgent(Agent):

    def echo(self, content):
        self.respondWith(content)

    def add(self, a, b):
        self.respondWith(a+b)
    
    def help(self):
        help = {
          "add": {
              "description": "add a and b",
              "args": ["aNumber", "aNumber"]
            },
          "echo": {
              "description": "echo the content",
              "args": ["aString"]
            }
          }
        self.respondWith({"help": help})



import json

from .agent import Agent, PyDemoAgent
from .space import MQTTSpace

class Supervisor:
    """
    I am responsible for looking after a group of agents (including managing their lifecycle). I am like a concierge. Whenever I receive a message from space, I will check the recipient of the message and deliver the message to that agent. If the agent wants to send out a message, it will give it to me first, and then I will forward it to the space.
    """

    def __init__(self) -> None:
        self.broadcastFlag = "[broadcast]"

        self.agents = {}
        # self.initAgents()
        self.space = MQTTSpace(self)
    
    def initAgents(self) -> None:
        agent = PyDemoAgent("PyDemoAgent")
        self.addAgent(agent)

    def addAgent(self, agent) -> Agent:
        agent.setSupervisor(self)
        self.agents[agent.id] = agent

    def getAgent(self, agentID) -> Agent:
        return self.agents[agentID]

    def onMessage(self, topic, payload) -> None:
        # route message to agents
        message = self.parseToJson(payload)
        if message:
            # log
            # print("(Supervisor) valid message: ", message)
            if message["to"] == self.broadcastFlag:
                for i in self.agents.values():
                    i._receive(message)

            if message["to"] in self.agents:
                self.agents[message["to"]]._receive(message)

    def parseToJson(self, payload):
        result = None
        try:
            result = json.loads(payload)

            # verify
            if self.isValid(result):
                return result
            else:
                print("(Supervisor) bad message")

        except Exception as e:
            print("(Supervisor) parseToJson error: " + str(e))

        return None

    def isValid(self, message):
        return ("from" in message and "to" in message and "action" in message)
    
    def send(self, message):
        routing_key = message["to"]
        self.space.publish(routing_key, json.dumps(message))


import logging
import asyncio
import os
import tempfile
from passlib.apps import custom_app_context

from amqtt.broker import Broker

# logger = logging.getLogger(__name__)


def generate_passwordfile():
    username = os.environ.get("MQTT_USERNAME", "guest")
    password = os.environ.get("MQTT_PASSWORD", "test")
    password_hash = custom_app_context.hash(password)

    temp_file = tempfile.NamedTemporaryFile(delete=False)

    # Write content to the temporary file
    with open(temp_file.name, "w") as file:
        file.write(f"{username}:{password_hash}")

    # Get the path of the temporary file
    temp_file_path = temp_file.name

    temp_file.close()

    return temp_file_path


config = {
    "listeners": {
        "default": {
            "type": "tcp",
            "bind": "0.0.0.0:1883",
        },
        "ws-mqtt": {
            "bind": "0.0.0.0:15675",
            "type": "ws",
            # "max_connections": 10,
        },
    },
    # "sys_interval": 10,
    "auth": {
        "allow-anonymous": False,
        "password-file": generate_passwordfile(),
        "plugins": ["auth_file", "auth_anonymous"],
    },
    "topic-check": {"enabled": False},
}

broker = Broker(config)

def main():
    formatter = "[%(asctime)s] :: %(levelname)s :: %(name)s :: %(message)s"
    formatter = "%(asctime)s :: %(levelname)s :: %(message)s"
    logging.basicConfig(level=logging.INFO, format=formatter)
    # loop = asyncio.new_event_loop() # to support thread
    # asyncio.set_event_loop(loop)
    asyncio.get_event_loop().run_until_complete(broker.start())
    asyncio.get_event_loop().run_forever()

if __name__ == "__main__":
    main()


# #!/usr/bin/env python

"""
# https://www.rabbitmq.com/tutorials/tutorial-five-python.html
# https://github.com/rabbitmq/rabbitmq-tutorials/blob/main/python/receive_logs_topic.py

Usage:

-    dynatalk-monitor  :  Subscribe to all message
-    dynatalk-monitor "+"  : Subscribe to all message
-    dynatalk-monitor --topics "SqueakDemoAgent"  : Subscribe to messages sent to SqueakDemoAgent
"""

import sys
import os
import time
import json
import argparse
import paho.mqtt.client as mqtt

from . import Supervisor, Agent


class MonitorSupervisor(Supervisor):

    def __init__(self, topics, json_only) -> None:
        super().__init__()
        self.topics = topics  # unused
        self.json_only = json_only

    def onMessage(self, topic, payload) -> None:
        # %r: raw string, repr()
        def set_string_color(text, color):
            reset_color = "\033[0m"
            color_map = {
                "green": "\033[32m",
                "red": "\033[31m",
                "blue": "\033[34m",
                "yellow": "\033[33m",
            }
            return color_map.get(color, reset_color) + text + reset_color

        # add timestamp, time when the message was received
        # import datetime
        # body_json["timestamp"] = datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S')
        message = self.parseToJson(payload)
        if not message:
            return
        
        format_body = json.dumps(message, sort_keys=True)

        # --json-only
        if self.json_only:
            print(format_body, flush=True)
        else:

            from_to = set_string_color(f'{message["from"]} -> {message["to"]}', "green")
            routing_key = f'routing_key : {message["to"]}'
            delimiter = set_string_color("|", "yellow")
            print(
                " [x] %s %s %s %s %s"
                % (from_to, delimiter, format_body, delimiter, routing_key),
                flush=True,
            )  # flush for linue pipline |


def _main(topics=None, json_only=False):

    if not topics:
        topics = ["+"]
    # print(topics, json_only)

    supervisor = MonitorSupervisor(topics, json_only)
    
    while True:
        time.sleep(1)

def monitor():
    parser = argparse.ArgumentParser(
        description="Python script with --topics and --json-only options."
    )
    parser.add_argument("--topics", nargs="*", help="List of topics.")
    parser.add_argument(
        "--json-only", action="store_true", default=False, help="Enable JSON-only mode."
    )
    args = parser.parse_args()
    if not args.json_only:
        # print(f'Dynatalk host: {os.getenv("MQTT_HOST")}', flush=True)
        pass
    try:
        _main(topics=args.topics, json_only=args.json_only)
    except KeyboardInterrupt:
        print("Interrupted")
        try:
            sys.exit(0)
        except SystemExit:
            os._exit(0)
